﻿-- Author      : RisM
-- Create Date : 11/28/2009 11:13:26 PM


local SpeakinSpell = LibStub("AceAddon-3.0"):GetAddon("SpeakinSpell")

SpeakinSpell:PrintLoading("network.lua")


-------------------------------------------------------------------------------
-- INVISIBLE NETWORK COMMUNICATIONS BETWEEN INSTANCES OF SPEAKINSPLL
-------------------------------------------------------------------------------
-- TODOFUTURE: add a global channel


function SpeakinSpell:Network_Init()
	if not SpeakinSpellSavedDataForAll.Networking.EnableNetwork then
		return
	end
		
	self:RegisterComm("SPEAKINSPELL", "Network_OnCommReceive")
	
	-- auto version check and other auto data syncs
	self:Network_AutoSync()
end


function SpeakinSpell:Network_Disable()
	self:UnregisterComm("SPEAKINSPELL")
end


-------------------------------------------------------------------------------
-- GENERAL COMMAND ROUTING - SEND/RECEIVE
-------------------------------------------------------------------------------



function SpeakinSpell:Network_GetBroadcastChannels()
	-- only use channels that apply to this character
	-- otherwise we could generate "you are not in a guild" spam, and the like
	local BroadcastChannels = {
		--"GUILD",
		--"RAID",
		--"PARTY",
	}
	if GetGuildInfo( UnitName("player") ) then
		tinsert(BroadcastChannels, "GUILD")
	end
	if GetNumPartyMembers() > 0 then
		tinsert(BroadcastChannels, "PARTY")
	end
	if GetNumRaidMembers() > 0 then
		tinsert(BroadcastChannels, "RAID")
	end
	return BroadcastChannels
end


-- send a command to all possible channels
function SpeakinSpell:Network_BroadcastCommand(command, data)
	local BroadcastChannels = self:Network_GetBroadcastChannels()
	for i,channel in pairs(BroadcastChannels) do
		self:Network_SendCommand(command, data, channel, nil)
	end
	
	-- if debugging is enabled, also whisper the broadcast message to myself
	if (SpeakinSpell.DEBUG_MODE or SpeakinSpellSavedData.ShowDebugMessages) then
		self:Network_SendCommand(command, data, "WHISPER", UnitName("player"))
	end
end


function SpeakinSpell:Network_SyncWithTarget(target)
	self:Network_SendCommand("V=", SpeakinSpell.CURRENT_VERSION, "WHISPER", target)
	self:Network_SendNewEvents(target, true, nil)
	self:Network_SendEventTable(target)
	self:Network_SendCommand("ET?", nil, "WHISPER", target)
end


function SpeakinSpell:Network_AutoSync() --TODONOW: ??call this function on joining a group??
	-- auto version check
	if SpeakinSpellSavedDataForAll.Networking.AutoVersionChecks then
		self:Network_SendVersionRequest()
	end
	-- auto sync NewEventsDetected
	if SpeakinSpellSavedDataForAll.Networking.AutoSyncNewEvents then
		self:Network_SendNewEvents(nil, true, nil)
	end
	-- auto sync EventTable
	if SpeakinSpellSavedDataForAll.Networking.AutoShareSpeeches then
		self:Network_SendEventTable(nil)
	end
	if SpeakinSpellSavedDataForAll.Networking.CollectSpeeches then
		self:Network_BroadcastCommand("ET?",nil)
	end
end


function SpeakinSpell:Network_SendCommand(command, data, distribution, target)
	local funcname = "Network_SendCommand"
	self:DebugMsg(funcname, string.format("distribution:%s, target:%s, command:%s", tostring(distribution), tostring(target), tostring(command)) )
	
	-- always pass: sender, clientversion, cmd, data
	local serializedData = self:Serialize( UnitName("player"), SpeakinSpell.CURRENT_VERSION, command, data )
	self:SendCommMessage("SPEAKINSPELL", serializedData, distribution, target)
end


function SpeakinSpell:Network_OnCommReceive(prefix, Msgs, distribution, target)
	local funcname = "Network_OnCommReceive"
	
	-- all messages are packaged in this form
	local success, sender, clientversion, cmd, data = self:Deserialize(Msgs)
	if not success then
		self:DebugMsg(funcname, "failed to deserialize" )
		return
	end
	
	self:DebugMsg(funcname, string.format("prefix:%s, distribution:%s, target:%s, from:%s, v:%s, cmd:%s", 
		tostring(prefix), 
		tostring(distribution), 
		tostring(target), 
		tostring(sender), 
		tostring(clientversion), 
		tostring(cmd)
		) )

	-- ignore messages from myself, unless debugging	
	if sender == UnitName("player") and not (SpeakinSpell.DEBUG_MODE or SpeakinSpellSavedData.ShowDebugMessages) then
		self:DebugMsg(funcname, "sender is self")
		return
	end
	
	-- route commands to functions
	local commands = {
		-- version number exchange
		["V="]	= function					( distribution, sender, clientversion, data )
			self:Network_OnVersion			( distribution, sender, clientversion, data )
		end,
		-- new events detected
		["NEW="]= function					( distribution, sender, clientversion, data )
			self:Network_OnNewEvents		( distribution, sender, clientversion, data )
		end,
		-- EventTable speech list sharing
		["ET="]	= function					( distribution, sender, clientversion, data )
			self:Network_OnEventTable		( distribution, sender, clientversion, data )
		end,
		["ET?"]	= function					( distribution, sender, clientversion, data )
			self:Network_OnEventTableRequest( distribution, sender, clientversion, data )
		end,
	}
	
	local func = commands[cmd]
	if not func then
		self:DebugMsg(funcname, "invalid cmd:"..tostring(cmd) )
		return
	end
	
	func( distribution, sender, clientversion, data )
end


-------------------------------------------------------------------------------
-- VERSION NUMBER EXCHANGE
-------------------------------------------------------------------------------


function SpeakinSpell:Network_SendVersionRequest()
	-- we're passive aggressive about this query
	-- broadcast what version we're running to every SS client we can
	-- any client that knows of a newer version will reply to us
	self:Network_BroadcastCommand("V=", SpeakinSpell.CURRENT_VERSION)
end


function SpeakinSpell:Network_OnVersion(distribution, sender, clientversion, data)
	local funcname = "Network_OnVersion"
	self:DebugMsg(funcname, tostring(data))
	
	-- make sure the data is in the expected format
	if type(data) ~= "string" then
		self:DebugMsg(funcname, "unexpected type(data) = "..type(data))
		return
	end
	
	if SpeakinSpell.DEBUG_MODE then
		self:DebugMsg(funcname, "suppressing broadcast of dev version")
		return
	end
	
	-- compare version numbers
	-- NOTE: NewestVersion is initialized to the current version
	--		but we can advertise for a newer client that may have gone offline
	
	if data > SpeakinSpell.RuntimeData.Networking.NewestVersion then
		-- an upgrade is available
		SpeakinSpell.RuntimeData.Networking.NewestVersion = data
		self:Print(L["Newer version available: "]..tostring(data))
		self:Print( SpeakinSpell.URL )
	
	elseif SpeakinSpell.RuntimeData.Networking.NewestVersion > data then
		-- we know of a newer version than this client, so tell them
		self:Network_SendCommand("V=", SpeakinSpell.RuntimeData.Networking.NewestVersion, "WHISPER", sender)
	end
end


-------------------------------------------------------------------------------
-- NEW EVENTS DETECTED
-------------------------------------------------------------------------------


function SpeakinSpell:Network_SendNewEvents(target, WantReply, MinusKnownEvents)
	-- send the list of new events detected with a flag to indicate whether we want a reply
	local data = {
		NewEvents = SpeakinSpellSavedDataForAll.NewEventsDetected,
		WantReply = WantReply
	}
	-- if a table of known events was provided, then don't send those to a client that already knows about them
	if MinusKnownEvents then
		data.NewEvents = {}
		for key,de in pairs(SpeakinSpellSavedDataForAll.NewEventsDetected) do
			local known = MinusKnownEvents[key]
			if not known then
				data.NewEvents[key] = de
			end
		end
		-- if there are no new events, then don't bother to reply
		if self:IsTableEmpty(data.NewEvents) then
			self:DebugMsg(funcname, "all events are already known to the far end client")
			return
		end
	end
	
	-- if a target is specified, send only to them, else broadcast to all
	if target then
		self:Network_SendCommand("NEW=", data, "WHISPER", target)
	else
		self:Network_BroadcastCommand("NEW=", data)
	end
end


function SpeakinSpell:Network_ValidateNewEvents( clientversion, data )
	local funcname = "Network_ValidateNewEvents"
	
	-- if this came from a client newer than us, then we must disregard it because we have no idea what could be new
	if clientversion > SpeakinSpell.CURRENT_VERSION then
		self:DebugMsg(funcname, "I'm afraid of data from a newer client v"..tostring(clientversion))
		return false
	end
	
	-- check for expected data types
	-- NewEvents = (table)SpeakinSpellSavedDataForAll.NewEventsDetected,
	-- WantReply = (bool)WantReply
	if type(data) ~= "table" then
		self:DebugMsg(funcname, "type(data):"..type(data))
		return false
	end
	if type(data.NewEvents) ~= "table" then
		self:DebugMsg(funcname, "type(data.NewEvents):"..type(data.NewEvents))
		return false
	end
	if type(data.WantReply) ~= "boolean" then
		self:DebugMsg(funcname, "type(data.WantReply):"..type(data.WantReply))
		return false
	end
	
	-- apply patches to the data if working with an older client
	if clientversion < SpeakinSpell.CURRENT_VERSION then
		self:PatchNetworkData_NewEventsDetected( clientversion, data.NewEvents )
	end
	
	-- NOTE: don't bother to Validate_DetectedEvent until we detect a new key
	return true
end


function SpeakinSpell:Network_OnNewEvents(distribution, sender, clientversion, data)
	local funcname = "Network_OnNewEvents"
	
	-- merge new events lists
	-- reply with anything new (non-redundant) in our data
	
	-- check for expected parameters
	-- and update data version format from older clients if necessary
	if not self:Network_ValidateNewEvents( clientversion, data ) then
		return
	end
		
	-- if the sender wants a reply, then we need to figure out what what new info we have available
	-- that's worthy of it's own subroutine
	if data.WantReply then
		self:Network_SendNewEvents(sender, false, data.NewEvents)
	end
	
	-- import the new events we just received
	for key,de in pairs(data.NewEvents) do
		local known = SpeakinSpellSavedDataForAll.NewEventsDetected[key]
		if not known then
			self:Validate_DetectedEvent(de)
			if key ~= de.key then
				self:DebugMsg(funcname, "validated de has key mismatch, disregarding it")
			else
				self:DebugMsg(funcname, "new key:"..tostring(key))
				SpeakinSpellSavedDataForAll.NewEventsDetected[key] = de
				-- force the GUI to update
				self.RuntimeData.OptionsGUIStates.CreateNew.FilterChanged = true
			end
		end
	end
end


-------------------------------------------------------------------------------
-- EVENT TABLES for CONTENT PACKS
-------------------------------------------------------------------------------


function SpeakinSpell:Network_SendEventTable(target)
	local funcname = "Network_SendEventTable"
	
	-- block sending speeches if privacy mode is enabled
	if not SpeakinSpellSavedDataForAll.Networking.ShareSpeeches then
		self:DebugMsg(funcname, "ShareSpeeches disabled")
		return
	end

	if target then
		self:Network_SendCommand("ET=", self:GetActiveEventTable(), "WHISPER", target)
	else
		self:Network_BroadcastCommand("ET=", self:GetActiveEventTable())
	end
end


function SpeakinSpell:Network_ValidateEventTable( clientversion, data )
	local funcname = "Network_ValidateEventTable"
	
	-- if this came from a client newer than us, then we must disregard it because we have no idea what could be new
	if tostring(clientversion) > SpeakinSpell.CURRENT_VERSION then
		self:DebugMsg(funcname, "I'm afraid of data from a newer client v"..tostring(clientversion))
		return false
	end
	
	-- check for expected data types
	-- data = EventTable
	if type(data) ~= "table" then
		self:DebugMsg(funcname, "type(data):"..type(data))
		return false
	end
	
	-- apply patches to the data if working with an older client
	if clientversion < SpeakinSpell.CURRENT_VERSION then
		self:PatchNetworkData_EventTable( clientversion, data )
	end
	
	-- also run validation functions
	-- new data added via default data structure templates sometimes do not get patch functions
	for key,ete in pairs( data ) do
		self:Validate_EventTableEntry(ete)
	end
	
	return true
end


function SpeakinSpell:Network_OnEventTable(distribution, sender, clientversion, data)
	local funcname = "Network_OnEventTable"
	
	-- disregard this if we don't collect speeches
	if not SpeakinSpellSavedDataForAll.Networking.CollectSpeeches then
		self:DebugMsg(funcname, "CollectSpeeches disabled")
		return
	end
	
	-- validate data
	if not self:Network_ValidateEventTable( clientversion, data ) then
		return
	end
	
	-- save the content for future use by the templates system and /ss import
	self:DebugMsg(funcname, "Collected EventTable from:"..tostring(sender))
	SpeakinSpellSavedDataForAll.Networking.CollectedEventTables[sender] = data
end


function SpeakinSpell:Network_OnEventTableRequest(distribution, sender, clientversion, data)
	if clientversion >= SpeakinSpell.CURRENT_VERSION then
		self:Network_SendEventTable(sender)
	end
end

